package eu.mhsl.craftattack.spawn.appliances.event;

import com.google.gson.Gson;
import eu.mhsl.craftattack.spawn.Main;
import eu.mhsl.craftattack.spawn.api.HttpServer;
import eu.mhsl.craftattack.spawn.appliance.Appliance;
import eu.mhsl.craftattack.spawn.appliance.ApplianceCommand;
import eu.mhsl.craftattack.spawn.appliances.event.command.EventCommand;
import eu.mhsl.craftattack.spawn.appliances.event.command.EventEndSessionCommand;
import eu.mhsl.craftattack.spawn.appliances.event.command.EventOpenSessionCommand;
import eu.mhsl.craftattack.spawn.appliances.event.command.MoveEventVillagerCommand;
import eu.mhsl.craftattack.spawn.appliances.event.listener.ApplyPendingRewardsListener;
import eu.mhsl.craftattack.spawn.util.entity.DisplayVillager;
import eu.mhsl.craftattack.spawn.util.server.PluginMessage;
import eu.mhsl.craftattack.spawn.util.listener.DismissInventoryOpenFromHolder;
import eu.mhsl.craftattack.spawn.util.listener.PlayerInteractAtEntityEventListener;
import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.format.NamedTextColor;
import org.bukkit.Bukkit;
import org.bukkit.Material;
import org.bukkit.entity.Player;
import org.bukkit.entity.Villager;
import org.bukkit.event.Listener;
import org.bukkit.inventory.ItemStack;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.http.HttpClient;
import java.net.http.HttpRequest;
import java.net.http.HttpResponse;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;

public class Event extends Appliance {
    public DisplayVillager.ConfigBound villager;
    private boolean isOpen = false;
    private UUID roomId;
    private final HttpClient eventServerClient = HttpClient.newHttpClient();
    private final List<Reward> pendingRewards = new ArrayList<>();

    record RewardConfiguration(String material, Map<UUID, Integer> rewards) {}
    record Reward(UUID playerUuid, Material material, int amount) {}

    public Event() {
        super("event");
    }

    @Override
    public void onEnable() {
        this.villager = new DisplayVillager.ConfigBound(
                localConfig(),
                villager -> {
                    villager.customName(Component.text("Events", NamedTextColor.GOLD));
                    villager.setProfession(Villager.Profession.LIBRARIAN);
                    villager.setVillagerType(Villager.Type.SNOW);
                }
        );
        this.isOpen = localConfig().getBoolean("enabled", false);
        if(this.isOpen) this.roomId = UUID.fromString(localConfig().getString("roomId", ""));
    }

    public void openEvent() throws URISyntaxException, IOException, InterruptedException {
        if(isOpen) throw new ApplianceCommand.Error("Es läuft derzeit bereits ein Event!");
        HttpRequest createRoomRequest = HttpRequest.newBuilder()
                .uri(new URI(localConfig().getString("api") + "/room"))
                .POST(HttpRequest.BodyPublishers.noBody())
                .build();

        HttpResponse<String> rawResponse = eventServerClient.send(createRoomRequest, HttpResponse.BodyHandlers.ofString());
        if(rawResponse.statusCode() != 200) throw new ApplianceCommand.Error("Event-Server meldet Fehler: " + rawResponse.statusCode());

        record Response(UUID uuid) {}
        Response response = new Gson().fromJson(rawResponse.body(), Response.class);

        isOpen = true;
        roomId = response.uuid;
    }

    public void joinEvent(Player p) {
        if(!isOpen) {
            p.sendMessage(Component.text("Zurzeit ist kein Event geöffnet.", NamedTextColor.RED));
            return;
        }

        try {
            p.sendMessage(Component.text("Authentifiziere...", NamedTextColor.GREEN));
            record Request(UUID player, UUID room) {}
            Request request = new Request(p.getUniqueId(), this.roomId);
            HttpRequest queueRoomRequest = HttpRequest.newBuilder()
                    .uri(new URI(localConfig().getString("api") + "/queueRoom"))
                    .POST(HttpRequest.BodyPublishers.ofString(new Gson().toJson(request)))
                    .build();

            HttpResponse<Void> rawResponse = eventServerClient.send(queueRoomRequest, HttpResponse.BodyHandlers.discarding());
            if(rawResponse.statusCode() != 200) {
                p.sendMessage(Component.text("Fehler beim beitreten: " + rawResponse.statusCode(), NamedTextColor.RED));
                return;
            }

            p.sendMessage(Component.text("Betrete...", NamedTextColor.GREEN));
            PluginMessage.connect(p, localConfig().getString("connect-server-name"));

        } catch (URISyntaxException | IOException | InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    public void endEvent() {
        if(!isOpen) throw new ApplianceCommand.Error("Es läuft derzeit kein Event!");
        isOpen = false;
    }

    private void rewardPlayers(RewardConfiguration rewardConfiguration) {
        rewardConfiguration.rewards.forEach((uuid, amount) -> {
            Reward reward = new Reward(uuid, Material.matchMaterial(rewardConfiguration.material), amount);
            if(reward.material == null) throw new RuntimeException("Material not found");

            if(Bukkit.getPlayer(uuid) == null) {
                pendingRewards.add(reward);
                return;
            }

            giveReward(reward);
        });
    }

    private void giveReward(Reward reward) {
        Player player = Bukkit.getPlayer(reward.playerUuid);
        if(player == null) throw new RuntimeException("Cannot reward offline playerUuid!");

        ItemStack rewardStack = new ItemStack(reward.material, reward.amount);
        Map<Integer, ItemStack> remaining = player.getInventory().addItem(rewardStack);
        Bukkit.getScheduler().runTask(
                Main.instance(),
                () -> remaining.values().forEach(remainingStack -> player.getWorld().dropItem(player.getLocation(), remainingStack))
        );
    }

    public void applyPendingRewards(Player player) {
        if(this.pendingRewards.isEmpty()) return;
        this.pendingRewards.stream()
                .filter(reward -> reward.playerUuid.equals(player.getUniqueId()))
                .forEach(this::giveReward);
    }

    @Override
    public void httpApi(HttpServer.ApiBuilder apiBuilder) {
        apiBuilder.post("reward", RewardConfiguration.class, (rewardConfiguration, request) -> {
            this.rewardPlayers(rewardConfiguration);
            return HttpServer.nothing;
        });
    }

    @Override
    protected @NotNull List<ApplianceCommand<?>> commands() {
        return List.of(
                new EventCommand(),
                new MoveEventVillagerCommand(),
                new EventOpenSessionCommand(),
                new EventEndSessionCommand()
        );
    }

    @Override
    protected @NotNull List<Listener> eventHandlers() {
        return List.of(
                new ApplyPendingRewardsListener(),
                new PlayerInteractAtEntityEventListener(this.villager.getUniqueId(), playerInteractAtEntityEvent -> joinEvent(playerInteractAtEntityEvent.getPlayer())),
                new DismissInventoryOpenFromHolder(this.villager.getUniqueId())
        );
    }
}
